/*
 * Part of network management sources. It is used in daemon that
 * checks which network devices are online on local subnets.
 *
 * Copyright (c) 2006 JAS
 *
 * Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
 *
 * $Id: ArpData.cpp 13 2006-04-27 01:36:19Z vokac $
 */
#include "ArpData.hpp"

#include <iostream>
#include <iomanip>
#include <sstream>

ArpData::ArpData()
{
  pthread_mutex_init(&_mutex, NULL);
}

ArpData::~ArpData() 
{
  clean();
  pthread_mutex_destroy(&_mutex);
}

unsigned int ArpData::convertIP(char *ip)
{
  if (ip == NULL)
    return 0;

  unsigned int ipc[4];
  if (sscanf(ip, "%u.%u.%u.%u", &ipc[0], &ipc[1], &ipc[2], &ipc[3]) != 4)
    return 0;

  return ((0xff & ipc[0]) << 24) + ((0xff & ipc[1]) << 16)
    + ((0xff & ipc[2]) << 8) + (0xff & ipc[3]);
}

unsigned char* ArpData::convertMAC(char *mac)
{
  if (mac == NULL)
    return NULL;

  unsigned int m[6];
  sscanf(mac, "%02x:%02x:%02x:%02x:%02x:%02x", &m[0], &m[1], &m[2], &m[3], &m[4], &m[5]);
  unsigned char *ret = new unsigned char[6];
  for (int i = 0; i < 6; i++)
    ret[i] = m[i];
  return ret;
}

char *ArpData::dump(unsigned int lower, unsigned int upper)
{
  ostringstream os;
  time_t tm = time(NULL);
  
  pthread_mutex_lock(&_mutex);

  for(_data_map::iterator i = _data.lower_bound(lower); i != _data.upper_bound(upper); i++) {
    DataInfo *data = i->second;
    unsigned int ip = data->ip;
    unsigned char *mac = data->mac;
    os << (ip>>24&0xff) << "." << (ip>>16&0xff) << "." << (ip>>8&0xff) << "." << (ip&0xff);
    os << " ";
    os.setf(ios::hex,ios::basefield); os.fill('0');
    os << setw(2) << int(mac[0]) << ":" << setw(2) << int(mac[1]) << ":";
    os << setw(2) << int(mac[2]) << ":" << setw(2) << int(mac[3]) << ":";
    os << setw(2) << int(mac[4]) << ":" << setw(2) << int(mac[5]);
    os.setf(ios::dec,ios::basefield); os.fill(' ');
    os << " " << tm - data->first;
    os << " " << tm - data->last;
    os << " " << tm - data->last_probe;
    os << endl;
  }

  pthread_mutex_unlock(&_mutex);

  return strdup(os.str().c_str());
}

int ArpData::add(unsigned int ip, const unsigned char mac[6], time_t tm)
{
  int retVal = UPDATE_RECORD_TIME;

  if (tm == 0)
    tm = time(NULL);

  pthread_mutex_lock(&_mutex);

  DataInfo *data = _data[ip];

  if (data == NULL) {
    data = (DataInfo *) malloc(sizeof(DataInfo));
    data->first = tm;
    data->last = tm;
    data->last_probe = tm;
    data->ip = ip;

    if (mac == NULL)
      memset(data->mac, 0, 6);
    else
      memcpy(data->mac, mac, 6);

    _data[ip] = data;

    retVal = NEW_RECORD;
  } else {
    data->last_probe = tm;

    if (mac != NULL) {
      data->last = tm;
      unsigned char *dmac = data->mac;

      if (dmac[0] != mac[0] || dmac[1] != mac[1]
	  || dmac[2] != mac[2] || dmac[3] != mac[3]
	  || dmac[4] != mac[4] || dmac[5] != mac[5]) {
	data->first = tm;
	memcpy(data->mac, mac, 6);
	
	retVal = UPDATE_RECORD_MAC;
      } else {
	retVal = UPDATE_RECORD_LAST;
      }
    }
  }
  
  pthread_mutex_unlock(&_mutex);

  return retVal;
}

int ArpData::add(const u_int8_t ip[4], const u_int8_t mac[6], time_t tm)
{
  return add((ip[0] << 24) + (ip[1] << 16) + (ip[2] << 8) + ip[3],
	     (const unsigned char *) mac, tm);
}

int ArpData::add(char *ip, char *mac, time_t tm)
{
  unsigned char *m = convertMAC(mac);
  int ret = add(convertIP(ip), m, tm);
  delete [] m;

  return ret;
}

void ArpData::clean(time_t from, time_t to)
{
  pthread_mutex_lock(&_mutex);
  for(_data_map::iterator i = _data.begin(); i != _data.end(); i++) {
    free(i->second);
  }
  _data.clear();
  pthread_mutex_unlock(&_mutex);
}

DataInfo* ArpData::list(time_t from, time_t to)
{
  DataInfo *data = (DataInfo *) malloc(sizeof(DataInfo)*_data.size());
  int n = 0;
  
  pthread_mutex_lock(&_mutex);
  for(_data_map::iterator i = _data.begin(); i != _data.end(); i++) {
    memcpy(&data[n], i->second, sizeof(DataInfo));
    n++;
  }
  pthread_mutex_unlock(&_mutex);

  return data;
}

vector<unsigned int>* ArpData::listIP(time_t from, time_t to)
{
  vector<unsigned int> *data = new vector<unsigned int>();
  data->reserve(_data.size());

  pthread_mutex_lock(&_mutex);
  for(_data_map::iterator i = _data.begin(); i != _data.end(); i++) {
    data->push_back(i->first);
  }
  pthread_mutex_unlock(&_mutex);

  return data;
}

DataInfo* ArpData::getByIP(unsigned int ip)
{
  DataInfo *data = (DataInfo *) malloc(sizeof(DataInfo));

  pthread_mutex_lock(&_mutex);
  DataInfo *ddata = _data[ip];
  if (ddata != NULL)
    memcpy(data, ddata, sizeof(DataInfo));
  pthread_mutex_unlock(&_mutex);

  return data;
}
  
DataInfo* ArpData::getByIP(char *ip)
{
  return getByIP(convertIP(ip));
}

DataInfo* ArpData::getByMAC(char mac[6])
{
  DataInfo *data = (DataInfo *) malloc(sizeof(DataInfo)*_data.size());
  int n = 0;
  
  pthread_mutex_lock(&_mutex);
  for(_data_map::iterator i = _data.begin(); i != _data.end(); i++) {
    unsigned char *dmac = i->second->mac;
    if (dmac[0] == mac[0] && dmac[1] == mac[1] && dmac[2] == mac[2]
	&& dmac[3] == mac[3] && dmac[4] == mac[4] && dmac[5] == mac[5]) {
      memcpy(&data[n], i->second, sizeof(DataInfo));
      n++;
    }
  }
  pthread_mutex_unlock(&_mutex);

  return data;
}
