/*
 * Part of network management sources. It is used in daemon that
 * checks which network devices are online on local subnets.
 *
 * Copyright (c) 2006 JAS
 *
 * Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
 *
 * $Id: Capture.cpp 7 2006-04-26 09:57:52Z vokac $
 */
#include "Capture.hpp"

#include <netinet/in.h>
#include <netinet/ether.h>

#define ETHERTYPE_8021Q 0x8100
#define ETHERTYPE_8021Q_SIZE 4


// constructor
// @param arp_data reference to ArpData object
Capture::Capture(ArpData *arp_data)
{
  syslog(LOG_INFO, "Capture::Capture(%p)", arp_data);

  pthread_mutex_init(&_capture_start_mutex, NULL);
  pthread_cond_init(&_capture_start_mutex_cond, NULL);

  _arp_data = arp_data;
}


// destructor
Capture::~Capture()
{
  syslog(LOG_INFO, "Capture::~Capture()");

  // stop pcap
  for (_data_map::iterator i = _data.begin(); i != _data.end(); i++) {
    const char *name = i->first;
    _pcap_thread *pt = i->second;
    pcap_t *p = pt->first;
    syslog(LOG_INFO, "stopping packet capture for interface %s", name);
    pcap_breakloop(p);
  }

  // stop/join threads
  for (_data_map::iterator i = _data.begin(); i != _data.end(); i++) {
    const char *name = i->first;
    _pcap_thread *pt = i->second;
    pthread_t *t = pt->second;
    syslog(LOG_INFO, "joining thread for interface %s", name);
    pthread_join(*t, NULL);
    delete t;
    pt->second = NULL;
  }

  // FIXME: how to delete has_map key??? free(name)

  // clean container
  _data.clear();

  pthread_mutex_destroy(&_capture_start_mutex);
  pthread_cond_destroy(&_capture_start_mutex_cond);
}


// Add new interface for listening network traffic
// @param name interface name
// @return success(0)/failure(-1)/warning(1)
int Capture::addInterface(const char *name)
{
  syslog(LOG_DEBUG, "Capture::addInterface(\"%s\")", name);
  if (_data.find(name) != _data.end()) {
    syslog(LOG_WARNING, "can't add interface that already exist: %s", name);
    return 1;
  }

  pthread_t *t = new pthread_t;

  // NOTE: this is ugly...
  capture_data *cd = new capture_data;
  cd->name = strdup(name);
  cd->pt = new _pcap_thread();
  cd->_capture_start_ok = &_capture_start_ok;
  cd->_capture_start_mutex = &_capture_start_mutex;
  cd->_capture_start_mutex_cond = &_capture_start_mutex_cond;
  cd->_arp_data = _arp_data;
  
  //syslog(LOG_ERR, "1aaa1: %s", name);
  pthread_mutex_lock(&_capture_start_mutex);
  _capture_start_ok = false;

  int rc = pthread_create(t, NULL, init_pcap_loop, (void *) cd);
  if (rc) {
    syslog(LOG_ERR, "return code from pthread_create() is %d", rc);
    pthread_mutex_unlock(&_capture_start_mutex);
    return -1;
  }
  cd->pt->second = t;

  pthread_cond_wait(&_capture_start_mutex_cond, &_capture_start_mutex);
  pthread_mutex_unlock(&_capture_start_mutex);

  if (!_capture_start_ok) {
    // capture startup failed
    syslog(LOG_ERR, "capture startup failed for interface %s, waiting to join and stop thread", name);
    pthread_join(*t, NULL);
    delete t;
    free(cd->name);
    delete cd->pt;
    return -1;
  }

  _data[cd->name] = cd->pt;
  delete cd;
  
  return 0;
}


// Stop listening netword traffic on defined interface
// @param name interface name
// @return success(0)/failure(-1)/warning(1)
int Capture::removeInterface(const char *name)
{
  syslog(LOG_DEBUG, "Capture::removeInterface(\"%s\")", name);
  if (_data.find(name) == _data.end()) {
    syslog(LOG_WARNING, "can't remove interface that doesn't exist: %s", name);
    return 1;
  }

  _pcap_thread *pt = _data.find(name)->second;

  // stop capturing packets, stop thread, free memory
  _data.erase(name);
  pcap_t *p = pt->first;
  syslog(LOG_INFO, "stopping packet capture for interface %s", name);
  pcap_breakloop(p);
  pthread_t *t = pt->second;
  syslog(LOG_INFO, "joining thread for interface %s", name);
  pthread_join(*t, NULL);
  free(t);

  return 0;
}


void *init_pcap_loop(void *data)
{
  char errbuf[PCAP_ERRBUF_SIZE];/* error buffer */
  pcap_t *handle;               /* packet capture handle */
  char filter_exp[] = "arp or rarp"; /* filter expression [3] */
  struct bpf_program fp;        /* compiled filter program (expression) */
  bpf_u_int32 mask = 0;         /* subnet mask */
  bpf_u_int32 net = 0;          /* ip */

  capture_data *cd = (capture_data *) data;
  char *name = cd->name;
  Capture::_pcap_thread *pt = cd->pt;
  bool *_capture_start_ok = cd->_capture_start_ok;
  pthread_mutex_t *_capture_start_mutex = cd->_capture_start_mutex;
  pthread_cond_t *_capture_start_mutex_cond = cd->_capture_start_mutex_cond;
  ArpData *_arp_data = cd->_arp_data;

  pthread_mutex_lock(_capture_start_mutex);
  syslog(LOG_INFO, "starting pcap for interface %s (%s)", name, filter_exp);

  /* get network number and mask associated with capture device */
  if (pcap_lookupnet(name, &net, &mask, errbuf) == -1) {
    syslog(LOG_NOTICE, "couldn't get netmask for device %s: %s", name, errbuf);
    net = 0;
    mask = 0;
  }
  syslog(LOG_DEBUG, "pcap net/mask for %s: %i.%i.%i.%i/%i.%i.%i.%i",
         name, net & 0xff, net >> 8 & 0xff, net >> 16 & 0xff,
         net >> 24 & 0xff, mask & 0xff, mask >> 8 & 0xff,
         mask >> 16 & 0xff, mask >> 24 & 0xff);

  /* open capture device */
  handle = pcap_open_live(name, Capture::maxPacketLength, 0, 1000, errbuf);
  if (handle == NULL) {
    syslog(LOG_ERR, "couldn't open device %s: %s", name, errbuf);
    pthread_cond_signal(_capture_start_mutex_cond);
    pthread_mutex_unlock(_capture_start_mutex);
    pthread_exit(NULL);
  }

  /* make sure we're capturing on an Ethernet device [2] */
  if (pcap_datalink(handle) != DLT_EN10MB) {
    syslog(LOG_ERR, "%s is not an Ethernet", name);
    pcap_close(handle);
    pthread_cond_signal(_capture_start_mutex_cond);
    pthread_mutex_unlock(_capture_start_mutex);
    pthread_exit(NULL);
  }

  /* compile the filter expression */
  if (pcap_compile(handle, &fp, filter_exp, 0, net) == -1) {
    syslog(LOG_ERR, "couldn't parse filter %s: %s", filter_exp, pcap_geterr(handle));
    pcap_close(handle);
    pthread_cond_signal(_capture_start_mutex_cond);
    pthread_mutex_unlock(_capture_start_mutex);
    pthread_exit(NULL);
  }

  /* apply the compiled filter */
  if (pcap_setfilter(handle, &fp) == -1) {
    syslog(LOG_ERR, "couldn't install filter %s: %s", filter_exp, pcap_geterr(handle));
    pcap_freecode(&fp);
    pcap_close(handle);
    pthread_cond_signal(_capture_start_mutex_cond);
    pthread_mutex_unlock(_capture_start_mutex);
    pthread_exit(NULL);
  }

  /* save capture handle, so we can access it from outside stop capture */
  pt->first = handle;

  /* send signal that we started correctly */
  *_capture_start_ok = true;
  pthread_cond_signal(_capture_start_mutex_cond);
  pthread_mutex_unlock(_capture_start_mutex);

  /* now we can set our callback function */
  syslog(LOG_INFO, "starting capture loop for interface %s", name);
  int npackets = pcap_loop(handle, 0, got_packet, (u_char *) _arp_data);
  syslog(LOG_DEBUG, "captured %d packets on interface %s", npackets, name);

  pt->first = NULL;

  /* cleanup */
  pcap_freecode(&fp);
  pcap_close(handle);

  pthread_exit(NULL);
}


void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
  static int count = 1;                   /* packet counter */

  ArpData *_arp_data = (ArpData *) args;
  /* declare pointers to packet headers */
  const struct ether_header *ether;
  const struct ether_arp *arp;

  syslog(LOG_DEBUG, "got_packet #%d [%d,%d]: ", count, header->caplen, header->len);
  count++;

  /* define ethernet and ARP/RARP header */
  ether = (const struct ether_header*)(packet);
  if (ntohs(ether->ether_type) == ETHERTYPE_ARP) {
    arp = (const struct ether_arp*)(packet + sizeof(struct ether_header));
  } else if (ntohs(ether->ether_type) == ETHERTYPE_8021Q) {
    arp = (const struct ether_arp*)(packet + sizeof(struct ether_header) + ETHERTYPE_8021Q_SIZE);
  } else {
    syslog(LOG_INFO, "unknown ether_type for ARP packet: %02x", ntohs(ether->ether_type));
    return;
  }

  syslog(LOG_DEBUG, "[%02x,%02x]: %02x:%02x:%02x:%02x:%02x:%02x - %02x:%02x:%02x:%02x:%02x:%02x",
	 ntohs(arp->arp_hrd), ntohs(arp->arp_pro),
	 ether->ether_shost[0], ether->ether_shost[1],
	 ether->ether_shost[2], ether->ether_shost[3],
	 ether->ether_shost[4], ether->ether_shost[5],
	 ether->ether_dhost[0], ether->ether_dhost[1],
	 ether->ether_dhost[2], ether->ether_dhost[3],
	 ether->ether_dhost[4], ether->ether_dhost[5]);

  if (ntohs(arp->arp_hrd) != ARPHRD_ETHER) return;
  if (ntohs(arp->arp_pro) != ETHERTYPE_IP) return;
  if (arp->arp_hln != 6) return;
  if (arp->arp_pln != 4) return;

  syslog(LOG_DEBUG, "[%02x,%d,%d]: %d.%d.%d.%d - %d.%d.%d.%d",
	 ntohs(arp->arp_op), arp->arp_hln, arp->arp_pln,
	 arp->arp_spa[0], arp->arp_spa[1],
	 arp->arp_spa[2], arp->arp_spa[3],
	 arp->arp_tpa[0], arp->arp_tpa[1],
	 arp->arp_tpa[2], arp->arp_tpa[3]);

  if (ntohs(arp->arp_op) == ARPOP_REQUEST or ntohs(arp->arp_op) == ARPOP_REPLY) {
    _arp_data->add(arp->arp_spa, arp->arp_sha);
//   } else if (ntohs(arp->arp_op) == ARPOP_REPLY) {
//     _arp_data->add(arp->arp_tpa, arp->arp_tha);
  } else {
    syslog(LOG_INFO, "unknown arpop: %02x", ntohs(arp->arp_op));
  }

  return;
}
