/*
 * Part of network management sources. It is used in daemon that
 * checks which network devices are online on local subnets.
 *
 * Copyright (c) 2006 JAS
 *
 * Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
 *
 * $Id: Command.cpp 13 2006-04-27 01:36:19Z vokac $
 */
#include "Command.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>


// constructor
// @param data reference to ArpData object
// @param req reference to Request object
Command::Command(command_conn_data *conn_data, ArpData *data, Request *req)
{
  syslog(LOG_DEBUG, "Command::Command(%p, %p, %p)", conn_data, data, req);

  _data = data;
  _req = req;

  _port = conn_data->port;
  _min_conn = conn_data->min_conn;
  _max_conn = conn_data->max_conn;

  // this must be called once in the program
  _ca_file = strdup(conn_data->ca_file);
  _crl_file = strdup(conn_data->crl_file);
  _cert_file = strdup(conn_data->cert_file);
  _key_file = strdup(conn_data->key_file);

  _user = strdup(conn_data->user);
  _pass = strdup(conn_data->pass);

  _ctx = initServerCTX();
  loadCertificates(_ctx, _cert_file, _key_file);

  struct sockaddr_in servaddr;

  /* Create a new stream (TCP) socket */
  _sock = socket(AF_INET, SOCK_STREAM, 0);
  if (_sock < 0) {
    syslog(LOG_ERR, "failed to create socket: %s", strerror(errno));
    throw new exception();
  }

  /* Enable address reuse */
  int on = 1;
  if (setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
    syslog(LOG_ERR, "set socket options failed: %s", strerror(errno));
    throw new exception();
  }
  
  /* Allow connections to port _port from any available interface */
  memset( &servaddr, 0, sizeof(servaddr) );
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl( INADDR_ANY );
  servaddr.sin_port = htons( _port );
  
  /* Bind to the address (interface/port) */
  if (bind(_sock, (struct sockaddr *) & servaddr, sizeof(servaddr)) < 0) {
    syslog(LOG_ERR, "bind to port %i failed: %s", _port, strerror(errno));
    throw new exception();
  }

  pthread_mutex_init(&_thread_pool_mutex, NULL);
}


// destructor
Command::~Command()
{
  syslog(LOG_DEBUG, "Command::~Command()");

  free(_ca_file);
  free(_crl_file);
  free(_cert_file);
  free(_key_file);

  free(_user);
  free(_pass);

  if (_ctx != NULL) SSL_CTX_free(_ctx);

  pthread_mutex_destroy(&_thread_pool_mutex);
}


// run command listener
void Command::run()
{
  syslog(LOG_DEBUG, "Command::run()");

  int rc, socket_id = 0;
  SSL *ssl;
  struct sockaddr_in cli_addr;
  socklen_t cli_len = sizeof(cli_addr);

  listen(_sock, _max_conn);
  my_listen_thread data_pool[_max_conn];
  memset(data_pool, 0, sizeof(my_listen_thread) * _max_conn);
  // for (int i = 0; i < _max_conn; i++) data_pool[i].in_use = false;

  while (1) {
    socket_id = accept(_sock, (struct sockaddr *) &cli_addr, &cli_len);

    if (socket_id < 0) {
      syslog(LOG_ERR, "acception connection failed: %s", strerror(errno));
      continue;
    }

    ssl = SSL_new(_ctx);
    if (SSL_set_fd(ssl, socket_id) == 0) {
      syslog(LOG_ERR, "error asociating socket with SSL: %s", ERR_error_string(ERR_get_error(), NULL));
      continue;
    }

    if (SSL_accept(ssl) != 1) {
      syslog(LOG_ERR, "handshake has failed: %s", ERR_error_string(ERR_get_error(), NULL));
      int sd = SSL_get_fd(ssl);
      SSL_free(ssl);
      close(sd);
      continue;
    }
    syslog(LOG_INFO, "handshake was completed");
    
    pthread_mutex_lock(&_thread_pool_mutex);

    int pos = 0;
    while (data_pool[pos].in_use) pos++;
    if (pos >= _max_conn) {
      syslog(LOG_ERR, "thread pool is full!?");
      int sd = SSL_get_fd(ssl);
      SSL_free(ssl);
      close(sd);
      pthread_mutex_unlock(&_thread_pool_mutex);
      break;
    }

    my_listen_thread *data = &data_pool[pos];
    data->in_use = true;
    data->socket_id = socket_id;
    data->ssl = ssl;
    data->_thread_pool_mutex = &_thread_pool_mutex;
    data->data = _data;
    data->req = _req;
    data->user = _user;
    data->pass = _pass;

    rc = pthread_create(&data->thread, NULL, connection, (void *) data);
    if (rc) {
      syslog(LOG_ERR, "return code from pthread_create() is %d", rc);
      data->in_use = false;
      int sd = SSL_get_fd(data->ssl);
      SSL_free(data->ssl);
      close(sd);
      pthread_mutex_unlock(&_thread_pool_mutex);
      break;
    }
    pthread_mutex_unlock(&_thread_pool_mutex);

  }

  for (int i = 0; i < _max_conn; i++) {
    if (data_pool[i].in_use) {
      pthread_mutex_lock(&_thread_pool_mutex);
      pthread_kill(data_pool[i].thread, 0);
      close(data_pool[i].socket_id);
      int sd = SSL_get_fd(data_pool[i].ssl);
      SSL_free(data_pool[i].ssl);
      close(sd);
      data_pool[i].in_use = false;
      //pthread_join?
      pthread_mutex_unlock(&_thread_pool_mutex);
    }
  }
}


SSL_CTX* Command::initServerCTX()
{
  SSL_METHOD *method;
  SSL_CTX *ctx;
  
  SSL_load_error_strings();
  SSL_library_init();
  method = SSLv23_method();
  ctx = SSL_CTX_new(method);
  if (ctx == NULL) {
    syslog(LOG_ERR, "error init CTX: %s", ERR_error_string(ERR_get_error(), NULL));
    throw new exception();
  }
  return ctx;
}


void Command::loadCertificates(SSL_CTX* ctx, char* CertFile, char* KeyFile)
{
  /* set the local certificate from CertFile */
  if ( SSL_CTX_use_certificate_file(ctx, CertFile, SSL_FILETYPE_PEM) <= 0 ) {
    syslog(LOG_ERR, "error loading certificate: %s", ERR_error_string(ERR_get_error(), NULL));
    throw new exception();
  }
  /* set the private key from KeyFile (may be the same as CertFile) */
  if ( SSL_CTX_use_PrivateKey_file(ctx, KeyFile, SSL_FILETYPE_PEM) <= 0 ) {
    syslog(LOG_ERR, "error loading private key: %s", ERR_error_string(ERR_get_error(), NULL));
    throw new exception();
  }
  /* verify private key */
  if ( !SSL_CTX_check_private_key(ctx) ) {
    syslog(LOG_ERR, "private key does not match the public certificate");
    throw new exception();
  }
}



void *connection(void* data)
{
  syslog(LOG_DEBUG, "connection(%p)", data);

  my_listen_thread *d = (my_listen_thread *) data;
  int status, offset = 0, buflen = 1024, quit = 0, pos, len;
  bool auth = false;
  char buffer[buflen+1], cmd[buflen], out[buflen];

  syslog(LOG_WARNING, "connection #%d ready to process commands", d->socket_id);

  while (!quit && (status = SSL_read(d->ssl, buffer+offset, buflen-offset)) > 0) {
    memset(buffer+offset+status, 0, buflen-offset-status+1);
    pos = 0;
    len = 0;
    syslog(LOG_DEBUG, "command[%i]: buffer[%i,%i] - %s", d->socket_id, offset, status, buffer);
    while (pos+len < buflen && buffer[pos+len] != 0) {
      while (pos+len < buflen && buffer[pos+len] != 0
	     && buffer[pos+len] != '\n' && buffer[pos+len] != '\r') len++;
      syslog(LOG_DEBUG, "command[%i]: [%d,%d] %s", d->socket_id, pos, len, buffer+pos);
      if (buffer[pos+len] == '\n' || buffer[pos+len] == '\r') {
	memcpy(cmd, buffer+pos, len);
	cmd[len] = 0;
	syslog(LOG_DEBUG, "command[%i]: %s[%d,%d]", d->socket_id, cmd, pos, len);
	if (strcasecmp(cmd, "ping") == 0) {
	  // PING - check if connection is valid
	  sprintf(out, "pong\n");
	  SSL_write(d->ssl, out, strlen(out));
	} else if (strcasecmp(cmd, "quit") == 0) {
	  // QUIT - client quit connection
	  quit = 1;
	  break;
	} else if (strncasecmp(cmd, "auth", 4) == 0) {
	  // AUTH - authenticate client with user/pass
	  if (strlen(cmd) < 6 or cmd[4] != ' ') {
	    sprintf(out, "undefined user\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else {
	    char* pom_user = cmd+5;
	    char* pom_pass = "";
	    char* pom_sep = strchr(pom_user, ' ');
	    if (pom_sep != NULL) {
	      pom_pass = pom_sep+1;
	      *pom_sep = '\0';
	    }
	    if (strcmp(d->user, pom_user) == 0 && strcmp(d->pass, pom_pass) == 0) {
	      auth = true;
	      sprintf(out, "ok\n");
	    } else {
	      sprintf(out, "fail %s/%s\n", pom_user, pom_pass);
	    }
	    syslog(LOG_INFO, "connection #%d auth: %s", d->socket_id, out);
	    SSL_write(d->ssl, out, strlen(out));
	  }
	} else if (strncasecmp(cmd, "dump", 4) == 0) {
	  // DUMP - debug data dump
	  unsigned int ip_from = 0;
	  unsigned int ip_to = (unsigned int) -1;
	  char* ip_pos = strchr(cmd, ' ');
	  if (ip_pos) {
	    ip_from = ArpData::convertIP(ip_pos+1);
	    ip_pos = strchr(ip_pos+1, ' ');
	    if (ip_pos) {
	      ip_to = ArpData::convertIP(ip_pos+1);
	      if (ip_to == 0)
		ip_to = ip_from;
	    } else
	      ip_to = ip_from;
	  }
	  char *dump_buffer = d->data->dump(ip_from, ip_to);
	  SSL_write(d->ssl, dump_buffer, strlen(dump_buffer));
	  free(dump_buffer);
	} else if (strncasecmp(cmd, "scan", 4) == 0) {
	  // SCAN - scan range of addresses
	  unsigned int ip_from = 0;
	  unsigned int ip_to = (unsigned int) -1;
	  bool ip_from_set = false;
	  bool ip_to_set = false;
	  unsigned int exp = 0, expNull = 0;
	  char *ip_pos = strchr(cmd, ' ');
	  if (ip_pos != NULL)
	    while (*ip_pos == ' ' and *ip_pos != '\0') ip_pos++;
	  if (ip_pos != NULL and *ip_pos != '\0') {
	    while (*ip_pos != '\0') {
	      while (*ip_pos == ' ' and *ip_pos != '\0') ip_pos++;
	      if (ip_pos[0] == '\0') break;
	      if (ip_pos[0] == '-') {
		ip_pos++;
		if (*ip_pos != '\0') {
		  char *ip_posp = ip_pos+1;
		  while (*ip_posp == ' ' and *ip_posp != '\0') ip_posp++;
		  char *ip_pospe = ip_posp;
		  while (*ip_pospe != ' ' and *ip_pospe != '\0') ip_pospe++;
		  if (*ip_pospe != '\0') {
		    *ip_pospe = '\0';
		    ip_pospe++;
		  }
		  if (ip_pos[0] == 'e') {
		    exp = atoi(ip_posp);
		  } else if (ip_pos[0] == 'n') {
		    expNull = atoi(ip_posp);
		  }
		  ip_pos = ip_pospe;
		}
	      } else {
		if (!ip_from_set) {
		  ip_from_set = true;
		  ip_from = ArpData::convertIP(ip_pos);
		} else if (!ip_to_set) {
		  ip_to_set = true;
		  ip_to = ArpData::convertIP(ip_pos);
		}
		while (*ip_pos != ' ' and *ip_pos != '\0') ip_pos++;
	      }
	    }
	  }
	  if (!ip_from_set and !ip_to_set) {
	    vector<unsigned int> *seenIP = d->data->listIP();
	    d->req->scan(seenIP, exp, expNull);
	    delete seenIP;
	  } else {
	    if (!ip_to_set)
	      ip_to = ip_from;
	    d->req->scan(ip_from, ip_to, exp, expNull);
	  }
	} else if (strncasecmp(cmd, "help", 4) == 0) {
	  // HELP - print command help
	  if (strncasecmp(cmd, "help auth", 9) == 0) {
	    sprintf(out, "auth user pass\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "  authenticate to gain access to the privileged functions\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else if (strncasecmp(cmd, "help dump", 9) == 0) {
	    sprintf(out, "dump [from_ip] [to_ip]\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "  dump data informations from ArpData\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else if (strncasecmp(cmd, "help help", 9) == 0) {
	    sprintf(out, "help [command]\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "  show help informations\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else if (strncasecmp(cmd, "help ping", 9) == 0) {
	    sprintf(out, "ping\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "  return pong, check if connection is valid\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else if (strncasecmp(cmd, "help quit", 9) == 0) {
	    sprintf(out, "quit\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "  close client connection\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else if (strncasecmp(cmd, "help scan", 9) == 0) {
	    sprintf(out, "scan [from_ip] [to_ip] [-e exp] [-n expNull]\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "  scan range of ip addresses\n");
	    SSL_write(d->ssl, out, strlen(out));
	  } else {
	    sprintf(out, "auth    authenticate user\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "dump    dump ArpData to stdout\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "help    this help, for details type \"help command\"\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "ping    pong\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "quit    close this connection\n");
	    SSL_write(d->ssl, out, strlen(out));
	    sprintf(out, "scan    snan addresses\n");
	    SSL_write(d->ssl, out, strlen(out));
	  }
	} else {
	  sprintf(out, "unknown command \"%s\", type \"help\" to get list of valid commands\n", cmd);
	  SSL_write(d->ssl, out, strlen(out));
	}
	sprintf(out, "\n");
	SSL_write(d->ssl, out, strlen(out));
      }
      if ((buffer[pos+len] == '\n' && buffer[pos+len+1] == '\r')
	  || (buffer[pos+len] == '\r' && buffer[pos+len+1] == '\n')) len++;
      pos += len + 1;
      len = 0;
    }
  }
  if (status == 0) {
    syslog(LOG_INFO, "peer has closed the SSL connection");
  } else if (status < 0) {
    syslog(LOG_WARNING, "received corrupted data(%d), closing the connection: %s", status, strerror(errno));
  }

  syslog(LOG_DEBUG, "closing command connection #%d", d->socket_id);

  pthread_mutex_lock(d->_thread_pool_mutex);
  int sd = SSL_get_fd(d->ssl);
  SSL_free(d->ssl);
  close(sd);
  d->in_use = false;
  pthread_mutex_unlock(d->_thread_pool_mutex);

  pthread_exit(NULL);
}
