/*
 * Part of network management sources. It is used in daemon that
 * checks which network devices are online on local subnets.
 *
 * Copyright (c) 2006 JAS
 *
 * Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
 *
 * $Id: Request.cpp 13 2006-04-27 01:36:19Z vokac $
 */
#include "Request.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/if_arp.h>


// constructor
Request::Request(ArpData *data)
{
  _data = data;

  struct in_addr src;

  _sock = socket(PF_PACKET, SOCK_DGRAM, 0);

  _me.sll_family = AF_PACKET;
  _me.sll_ifindex = 2; // FIXME: !!!!!!! ifindex;
  _me.sll_protocol = htons(ETH_P_ARP);
  
  if (bind(_sock, (struct sockaddr*)&_me, sizeof(_me)) == -1) {
    syslog(LOG_ERR, "bind failed: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }

  int alen = sizeof(_me);
  if (getsockname(_sock, (struct sockaddr*)&_me, (socklen_t*)&alen) == -1) {
    syslog(LOG_ERR, "getsockname failed: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }

  _he = _me;
  //  memcpy(&he, &me, sizeof(me));
  memset(_he.sll_addr, -1, (_he.sll_halen < sizeof(_he.sll_addr) ? _he.sll_halen : sizeof(_he.sll_addr)));

  struct ifreq ifreq;
  ifreq.ifr_ifindex = _me.sll_ifindex;
  if (ioctl(_sock, SIOCGIFNAME, &ifreq) < 0) {
    syslog(LOG_ERR, "unable to get interface name %i: %s", ifreq.ifr_ifindex, strerror(errno));
    exit(EXIT_FAILURE);
  }
  if (ioctl(_sock, SIOCGIFADDR, &ifreq) < 0) {
    syslog(LOG_ERR, "unable to get address for %s: %s", ifreq.ifr_name, strerror(errno));
    exit(EXIT_FAILURE);
  }
  src = ((struct sockaddr_in *) &ifreq.ifr_addr)->sin_addr;
  syslog(LOG_INFO, "arp source address %s: %s", ifreq.ifr_name, inet_ntoa(src));
  
  struct arphdr *ah = (struct arphdr*)_buf;
  unsigned char *p = (unsigned char *)(ah+1);
  
  ah->ar_hrd = htons(ARPHRD_ETHER);
  ah->ar_pro = htons(ETH_P_IP);
  ah->ar_hln = 6;
  ah->ar_pln = 4;
  ah->ar_op  = htons(ARPOP_REQUEST);

  memcpy(p, _me.sll_addr, ah->ar_hln);
  p += ah->ar_hln;
  memcpy(p, &src, 4);
  p += 4;
  
  memcpy(p, _he.sll_addr, ah->ar_hln);
  p += ah->ar_hln;
  memcpy(p, &src, 4); // will be overwrited
  _buf_pos = p;
  p += 4;

  _buf_len = p - _buf;
}


// destructor
Request::~Request()
{
  close(_sock);
}


// send new arp request for ip/name
void Request::send(unsigned int ip, int exp, int expNull)
{
  DataInfo *dataInfo = _data->getByIP(ip);
  if (dataInfo != NULL) {
    bool exitEarly = false;

    if (exp > 0) {
      time_t tm = time(NULL);
      unsigned char *macNull = dataInfo->mac;
      
      if (expNull == 0) expNull = exp;
      if (macNull[0] == 0 && macNull[1] == 0 && macNull[2] == 0
	  && macNull[3] == 0 && macNull[4] == 0 && macNull[5] == 0) {
	if (dataInfo->last_probe - tm <= expNull) exitEarly = true;
      } else {
	if (dataInfo->last_probe - tm <= exp) exitEarly = true;
      }
    }
    
    free(dataInfo);

    if (exitEarly) return;
  }

  memcpy(_buf_pos, (struct in_addr *) &ip, 4);
  int rc = sendto(_sock, _buf, _buf_len, 0, (struct sockaddr*)&_he, sizeof(struct sockaddr_ll));

  syslog(LOG_DEBUG, "scaning %s: %i (%i: %s)",
	 inet_ntoa(inet_makeaddr(ip, ip)), rc, errno, strerror(errno));

  if (rc > 0)
    _data->add(ip, NULL);
}


// send new arp request for range IP addresses
// @param from_ip lower bound of IP addresses
// @param to_ip upper bound of IP addresses
// @param tolerance time to determine "expired" records
void Request::scan(unsigned int from_ip, unsigned int to_ip, int exp, int expNull)
{
  syslog(LOG_INFO, "scan(%s, %s, %i, %i)",
	 inet_ntoa(inet_makeaddr(from_ip, from_ip)),
	 inet_ntoa(inet_makeaddr(to_ip, to_ip)),
	 exp, expNull);
  
  for (unsigned int ip = from_ip; ip <= to_ip; ip++) {
    send(ip);
  }
}

// send new arp request for list of IP addresses
// @param ip vector of IP addresses to scan
// @param tolerance time to determine "expired" records
void Request::scan(vector<unsigned int> *ip, int exp, int expNull)
{
  syslog(LOG_INFO, "scan(vector<%i>, %i, %i)", (int) ip->size(), exp, expNull);

  for (vector<unsigned int>::iterator i = ip->begin(); i != ip->end(); i++) {
    send(*i, exp, expNull);
  }
}
