/*
 * Part of network management sources. It is used in daemon that
 * checks which network devices are online on local subnets.
 *
 * Copyright (c) 2006 JAS
 *
 * Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
 *
 * $Id: arpsniff.c 13 2006-04-27 01:36:19Z vokac $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <syslog.h>
#include <pcap.h>

#include "Capture.hpp"
#include "Request.hpp"
#include "Command.hpp"

#define DEF_INTERFACE NULL
#define DEF_PORT 9320

/* Flag set by `--verbose'. */
static int verbose_flag;

void help()
{
  fprintf(stdout, "Parameters:\n");
  fprintf(stdout, "  -i, --interface    comma separated list of network interfaces (eth0,eth1,...)\n");
  fprintf(stdout, "                     by default listen on all interfaces\n");
  fprintf(stdout, "  -d, --daemon       daemon mode\n");
  fprintf(stdout, "  -p, --port         TCP port for command connection\n");
  fprintf(stdout, "  -u, --user         auth username\n");
  fprintf(stdout, "  -a, --password     auth password\n");
  fprintf(stdout, "  --ca-file          file with CA certs\n");
  fprintf(stdout, "  --crl-file         file with CA crls\n");
  fprintf(stdout, "  --cert-file        file with server cert\n");
  fprintf(stdout, "  --key-file         file with server key\n");
  fprintf(stdout, "  -h, --help         this help\n");
  fprintf(stdout, "  -q, --quite        quite mode\n");
  fprintf(stdout, "  -v, --verbose      verbose mode\n");
}


void usage(char *prg)
{
  fprintf(stdout, "ARPSniff $Revision: 13 $ ( http://kmlinux.fjfi.cvut.cz/~vokac/activities/arpsniff )\n");
  fprintf(stdout, "Usage: %s [-i int1,int2] [-d] [-p port] [-h] [-q] [-v]\n", prg);
}

void daemonize()
{
  pid_t pid, sid;
        
  /* Fork off the parent process */
  pid = fork();
  if (pid < 0) {
    syslog(LOG_ERR, "forking daemon process failed: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }
  /* If we got a good PID, then
     we can exit the parent process. */
  if (pid > 0) {
    syslog(LOG_DEBUG, "exiting parent process");
    exit(EXIT_SUCCESS);
  }
  
  /* Change the file mode mask */
  umask(0);       
  
  /* Open any logs here */
  
  /* Create a new SID for the child process */
  sid = setsid();
  if (sid < 0) {
    /* Log any failure here */
    syslog(LOG_DEBUG, "can't create new SID: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  /* Change the current working directory */
  if ((chdir("/")) < 0) {
    /* Log any failure here */
    syslog(LOG_DEBUG, "can't chagne working directory: %s", strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  
  /* Close out the standard file descriptors */
  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);
}

int main (int argc, char **argv)
{
  int c;
  int daemon = 0;
  char *interface = DEF_INTERFACE;
  
  // start logging first
  //  setlogmask(LOG_UPTO(LOG_NOTICE));
  setlogmask(LOG_UPTO(LOG_DEBUG));
  openlog("arpsniff", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

  command_conn_data conn_data;
  conn_data.port = DEF_PORT;
  conn_data.min_conn = 1;
  conn_data.max_conn = 10;
  conn_data.ca_file = "/etc/pki/tls/certs/ca-bundle.crt";
  conn_data.crl_file = "/etc/pki/tls/certs/ca-bundle.crl";
  conn_data.cert_file = "/etc/pki/tls/certs/localhost.crt";
  conn_data.key_file = "/etc/pki/tls/private/localhost.key";
  conn_data.user = "";
  conn_data.pass = "";

  while (1) {

    static struct option long_options[] = {
      /* These options set a flag. */
      {"verbose",     no_argument,       &verbose_flag, 1},
      {"quite",       no_argument,       &verbose_flag, 0},
      /* These options don't set a flag.
	 We distinguish them by their indices. */
      {"daemon",      no_argument,       0, 'd'},
      {"interface",   no_argument,       0, 'i'},
      {"port",        required_argument, 0, 'p'},
      {"user",        required_argument, 0, 'u'},
      {"password",    required_argument, 0, 'a'},
      {"ca-file",     required_argument, 0, 0},
      {"crl-file",    required_argument, 0, 0},
      {"cert-file",   required_argument, 0, 0},
      {"key-file",    required_argument, 0, 0},
      {"help",        no_argument,       0, 'h'},
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index = 0;
    
    c = getopt_long (argc, argv, "di:c:e:hvq",
		     long_options, &option_index);
    
    /* Detect the end of the options. */
    if (c == -1)
      break;
    
    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
	break;
      if (strcmp(long_options[option_index].name, "ca-file") == 0) {
	conn_data.ca_file = strdup(optarg);
      } else if (strcmp(long_options[option_index].name, "crl-file") == 0) {
	conn_data.crl_file = strdup(optarg);
      } else if (strcmp(long_options[option_index].name, "cert-file") == 0) {
	conn_data.cert_file = strdup(optarg);
      } else if (strcmp(long_options[option_index].name, "key-file") == 0) {
	conn_data.key_file = strdup(optarg);
      } else {
	printf("option %s", long_options[option_index].name);
	if (optarg)
	  printf(" with arg %s", optarg);
	printf("\n");
      }
      break;
      
    case 'd':
      daemon = 1;
      break;
      
    case 'i':
      interface = optarg;
      break;
      
    case 'p':
      conn_data.port = atoi(optarg);
      break;
      
    case 'u':
      conn_data.user = strdup(optarg);
      break;
      
    case 'a':
      conn_data.pass = strdup(optarg);
      break;
      
    case 'h':
      usage(argv[0]);
      help();
      exit(EXIT_SUCCESS);
      break;
      
    case '?':
      /* getopt_long already printed an error message. */
      usage(argv[0]);
      help();
      exit(EXIT_SUCCESS);
      break;
     
    default:
      abort ();
    }
  }
  
  /* Print any remaining command line arguments (not options). */
  if (verbose_flag && optind < argc) {
    while (optind < argc)
      syslog(LOG_INFO, "non-option ARGV-elements: %s", argv[optind++]);
  }
  
  /* check user privileges - we need root privileges */
  if (geteuid() != 0) {
    fprintf(stderr, "you need root privileges to capture network traffic\n");
    exit(EXIT_FAILURE);
  }

  if (daemon)
    daemonize();

  /* create threads for capturing APR requests/responses */
  ArpData *data = new ArpData();
  Capture *cap = new Capture(data);
  if (interface != NULL) {
    // use interface specified on command line
    char *copy_interface = strdup(interface);
    char *token;
    token = strtok(copy_interface, ",");
    while (token != NULL) {
      if (verbose_flag) printf("%s\n", token);
      cap->addInterface(token);
      token = strtok(NULL, ",");
    }
    free(copy_interface);
  } else {
    // use interfaces found by pcap function pcap_findalldevs
    char errbuf[PCAP_ERRBUF_SIZE];
    pcap_if_t *alldevs, *dev;
    if (pcap_findalldevs(&alldevs, errbuf) < 0) {
      fprintf(stderr, "error searching all network interfaces: %s\n", errbuf);
      exit(EXIT_FAILURE);
    }
    dev = alldevs;
    while (dev != NULL) {
      if (dev->flags != PCAP_IF_LOOPBACK and strcmp(dev->name, "any") != 0) {
	if (verbose_flag) printf("%s\n", dev->name);
	cap->addInterface(dev->name);
      }
      dev = dev->next;
    }
    pcap_freealldevs(alldevs);
  }
  
  Request *req = new Request(data);
  Command *cmd = new Command(&conn_data, data, req);
  cmd->run();
  delete cmd;
  delete req;
  delete cap;
  delete data;

  // close logging
  closelog();

  exit(EXIT_SUCCESS);
}
